﻿using System;
using System.IO;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;

class Program
{
    [DllImport("kernel32.dll", SetLastError = true, CharSet = CharSet.Auto)]
    [return: MarshalAs(UnmanagedType.Bool)]
    unsafe static extern bool CopyFileEx(string lpExistingFileName, string lpNewFileName,
            delegate* unmanaged[Stdcall]<long, long, long, long, uint, CopyProgressCallbackReason, IntPtr, IntPtr, IntPtr, CopyProgressResult> lpProgressRoutine,
            IntPtr lpData, ref Int32 pbCancel, CopyFileFlags dwCopyFlags);

    [UnmanagedCallersOnly(CallConvs = new Type[] { typeof(CallConvStdcall) })]
    static CopyProgressResult CopyProgressRoutineCallback(
        long TotalFileSize,
        long TotalBytesTransferred,
        long StreamSize,
        long StreamBytesTransferred,
        uint dwStreamNumber,
        CopyProgressCallbackReason dwCallbackReason,
        IntPtr hSourceFile,
        IntPtr hDestinationFile,
        IntPtr lpData)
    {
        long n = TotalFileSize;
        Console.WriteLine($"{TotalBytesTransferred} / {TotalFileSize}");
        return CopyProgressResult.PROGRESS_CONTINUE;
    }

    static void Main(string[] args)
    {
        string srcFile = @"...[large-sized-file-path]...";
        string dstFile = Path.ChangeExtension(srcFile, ".tmp");

        if (IntPtr.Size != 8)
        {
            throw new NotSupportedException("\"Function pointer + UnmanagedCallersOnly with calling conventions\" are not supported in 32-bit mode");
        }

        int bCancel = 0;
        unsafe
        {
            CopyFileEx(srcFile, dstFile, &CopyProgressRoutineCallback, IntPtr.Zero, ref bCancel, CopyFileFlags.COPY_FILE_RESTARTABLE);
        }

        Console.WriteLine("Press any key to exit...");
        Console.ReadLine();
    }
}

[Flags]
enum CopyFileFlags : uint
{
    COPY_FILE_NONE = 0,
    COPY_FILE_FAIL_IF_EXISTS = 0x00000001,
    COPY_FILE_RESTARTABLE = 0x00000002,
    COPY_FILE_OPEN_SOURCE_FOR_WRITE = 0x00000004,
    COPY_FILE_ALLOW_DECRYPTED_DESTINATION = 0x00000008,
    COPY_FILE_COPY_SYMLINK = 0x00000800, //NT 6.0+
    COPY_FILE_NO_BUFFERING = 0x00001000 //NT 6.0+
}

enum CopyProgressCallbackReason : uint
{
    CALLBACK_CHUNK_FINISHED = 0x00000000,
    CALLBACK_STREAM_SWITCH = 0x00000001
}

enum CopyProgressResult : uint
{
    PROGRESS_CONTINUE = 0,
    PROGRESS_CANCEL = 1,
    PROGRESS_STOP = 2,
    PROGRESS_QUIET = 3
}

#if !NET5_0
namespace System.Runtime.InteropServices
{
    [AttributeUsage(AttributeTargets.Method, Inherited = false)]
    public sealed class UnmanagedCallersOnlyAttribute : Attribute
    {
        public Type[] CallConvs;
        public string EntryPoint;
    }
}
#endif